//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// MLEMutator.cc -- 
//
// Author           : Rodolfo G. Esteves
// Created On       : Mon May 18 07:44:20 2015
// Last Modified By : Rob Schluntz
// Last Modified On : Tue Jun 02 13:43:01 2015
// Update Count     : 92
//

#include <cassert>
#include <algorithm>

#include "MLEMutator.h"
#include "SynTree/Statement.h"

namespace ControlStruct {
	MLEMutator::~MLEMutator() {
		delete targetTable;
		targetTable = 0;
	}

	CompoundStmt* MLEMutator::mutate( CompoundStmt *cmpndStmt ) {
		bool labeledBlock = false;
		if ( !(cmpndStmt->get_labels().empty()) ) {
			labeledBlock = true;
			enclosingBlocks.push_back( Entry( cmpndStmt ) );
		} // if

		// a child statement may set the break label
		// - if they do, attach it to the next statement
		std::list< Statement * > &kids = cmpndStmt->get_kids();
		for ( std::list< Statement * >::iterator k = kids.begin(); k != kids.end(); k++ ) {
			*k = (*k)->acceptMutator(*this);

			if ( ! get_breakLabel().empty() ) {
				std::list< Statement * >::iterator next = k+1;
				if ( next == kids.end() ) {
					std::list<Label> ls; ls.push_back( get_breakLabel() );
					kids.push_back( new NullStmt( ls ) );
				} else {
					(*next)->get_labels().push_back( get_breakLabel() );
				}

				set_breakLabel("");
			} // if
		} // for

		if ( labeledBlock ) {
			assert( ! enclosingBlocks.empty() );
			if ( ! enclosingBlocks.back().get_breakExit().empty() ) {
				set_breakLabel( enclosingBlocks.back().get_breakExit() );
			}
			enclosingBlocks.pop_back();
		} // if

		return cmpndStmt;
	}

	Statement *MLEMutator::mutate( WhileStmt *whileStmt ) {
		enclosingLoops.push_back( Entry( whileStmt ) );
		whileStmt->set_body ( whileStmt->get_body()->acceptMutator( *this ) );

		Entry &e = enclosingLoops.back();
		assert ( e == whileStmt );
		whileStmt->set_body( mutateLoop( whileStmt->get_body(), e ) );
		enclosingLoops.pop_back();

		return whileStmt;
	}

	Statement *MLEMutator::mutate( ForStmt *forStmt ) {
		enclosingLoops.push_back( Entry( forStmt ) );
		forStmt->set_body( maybeMutate( forStmt->get_body(), *this ) );

		Entry &e = enclosingLoops.back();
		assert ( e == forStmt );
		forStmt->set_body( mutateLoop( forStmt->get_body(), e ) );
		enclosingLoops.pop_back();

		return forStmt;
	}

	Statement *MLEMutator::mutate( BranchStmt *branchStmt ) throw ( SemanticError ) {
		std::string originalTarget = branchStmt->get_originalTarget();

		if ( branchStmt->get_type() == BranchStmt::Goto )
			return branchStmt;

		// test if continue target is a loop
		if ( branchStmt->get_type() == BranchStmt::Continue) {
			if ( enclosingLoops.empty() ) {
				throw SemanticError( "'continue' outside a loop" );
			} else if ( std::find( enclosingLoops.begin(), enclosingLoops.end(), (*targetTable)[branchStmt->get_target()] ) == enclosingLoops.end() ) {
				throw SemanticError( "'continue' target label must be an enclosing loop: " + originalTarget );
			}
		}

		if ( branchStmt->get_type() == BranchStmt::Break && (enclosingLoops.empty() && enclosingSwitches.empty() && enclosingBlocks.empty() ) )
			throw SemanticError( "'break' outside a loop or switch" );

		if ( branchStmt->get_target() == "" ) return branchStmt;

		if ( targetTable->find( branchStmt->get_target() ) == targetTable->end() )
			throw SemanticError("The label defined in the exit loop statement does not exist: " + originalTarget );  // shouldn't happen (since that's already checked)

		std::list< Entry >::iterator check;
		if ( ( check = std::find( enclosingLoops.begin(), enclosingLoops.end(), (*targetTable)[branchStmt->get_target()] ) ) == enclosingLoops.end() )
			// not in loop, checking if in block
			if ( (check = std::find( enclosingBlocks.begin(), enclosingBlocks.end(), (*targetTable)[branchStmt->get_target()] )) == enclosingBlocks.end() )
				// neither in loop nor in block, checking if in switch/choose
				if ( (check = std::find( enclosingSwitches.begin(), enclosingSwitches.end(), (*targetTable)[branchStmt->get_target()] )) == enclosingSwitches.end() )
					throw SemanticError("The target specified in the exit loop statement does not correspond to an enclosing control structure: " + originalTarget );

		if ( enclosingLoops.back() == (*check) )
			return branchStmt;				// exit the innermost loop (labels unnecessary)

		Label newLabel;
		switch ( branchStmt->get_type() ) {
		  case BranchStmt::Break:
				if ( check->get_breakExit() != "" ) {
					newLabel = check->get_breakExit();
				} else {
					newLabel = generator->newLabel();
					check->set_breakExit( newLabel );
				} // if
				break;
		  case BranchStmt::Continue:
				if ( check->get_contExit() != "" ) {
					newLabel = check->get_contExit();
				} else {
					newLabel = generator->newLabel();
					check->set_contExit( newLabel );
				} // if
				break;
		  default:
				return 0;					// shouldn't be here
		} // switch

		return new BranchStmt( std::list<Label>(), newLabel, BranchStmt::Goto );
	}

	template< typename SwitchClass >
	Statement *handleSwitchStmt( SwitchClass *switchStmt, MLEMutator &mutator ) {
		// set up some aliases so that the rest of the code isn't messy
		typedef MLEMutator::Entry Entry;
		LabelGenerator *generator = mutator.generator;
		std::list< Entry > &enclosingSwitches = mutator.enclosingSwitches;

		Label brkLabel = generator->newLabel();
		enclosingSwitches.push_back( Entry(switchStmt, "", brkLabel) );
		mutateAll( switchStmt->get_branches(), mutator ); 
		{
			// check if this is necessary (if there is a break to this point, otherwise do not generate

			// for the purposes of keeping switch statements uniform (i.e. all statements that are 
			// direct children of a switch should be CastStmts), append the exit label + break to the 
			// last case statement; create a default case if there are no cases
			std::list< Statement * > &branches = switchStmt->get_branches();
			if ( branches.empty() ) {
				branches.push_back( CaseStmt::makeDefault() );
			}

			if ( CaseStmt * c = dynamic_cast< CaseStmt * >( branches.back() ) ) {
				std::list<Label> temp; temp.push_back( brkLabel );
				c->get_statements().push_back( new BranchStmt( temp, Label(""), BranchStmt::Break ) );
			} else assert(0);
		}
		assert ( enclosingSwitches.back() == switchStmt );
		enclosingSwitches.pop_back();
		return switchStmt;
	}

	Statement *MLEMutator::mutate( SwitchStmt *switchStmt ) {
		return handleSwitchStmt( switchStmt, *this );
	}

	Statement *MLEMutator::mutate( ChooseStmt *switchStmt ) {
		return handleSwitchStmt( switchStmt, *this );		
	}

	Statement *MLEMutator::mutateLoop( Statement *bodyLoop, Entry &e ) {
		// ensure loop body is a block
		CompoundStmt *newBody;
		if ( ! (newBody = dynamic_cast<CompoundStmt *>( bodyLoop )) ) {
			newBody = new CompoundStmt( std::list< Label >() );
			newBody->get_kids().push_back( bodyLoop );
		} // if

		Label endLabel = e.get_contExit();

		if ( e.get_breakExit() != "" ) {
			if ( endLabel == "" ) endLabel = generator->newLabel();
			// check for whether this label is used or not, so as to not generate extraneous gotos
			if (e.breakExitUsed)
				newBody->get_kids().push_back( new BranchStmt( std::list< Label >(), endLabel, BranchStmt::Goto ) );
			// xxx
			//std::list< Label > ls; ls.push_back( e.get_breakExit() );
			set_breakLabel( e.get_breakExit() );
			//newBody->get_kids().push_back( new BranchStmt( ls, "", BranchStmt::Break ) );
		} // if

		if ( e.get_breakExit() != "" || e.get_contExit() != "" ) {
			if (dynamic_cast< NullStmt *>( newBody->get_kids().back() ))
				newBody->get_kids().back()->get_labels().push_back( endLabel );
			else {
				std::list < Label > ls; ls.push_back( endLabel );
				newBody->get_kids().push_back( new NullStmt( ls ) );
			} // if
		} // if

		return newBody;
	}

	//*** Entry's methods - ensure these labels can be set at most once
	void MLEMutator::Entry::set_contExit( Label l ) {
		assert ( contExit == "" || contExit == l );
		contExit = l;
	}

	void MLEMutator::Entry::set_breakExit( Label l ) {
		assert ( breakExit == "" || breakExit == l );
		breakExit = l;
	}
} // namespace ControlStruct

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
